#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
# https://cortix.org
"""
Cortix Example Module
"""

import logging
from copy import deepcopy

import scipy.constants as unit
import numpy as math

import iapws.iapws97 as steam_table

from cortix import Module
from cortix.support.phase_new import PhaseNew as Phase
from cortix import Quantity

class Turbine(Module):
    """Boiling water single-point reactor.

    Notes
    -----
    These are the `port` names available in this module to connect to respective
    modules: `turbine`, and `pump`.
    See instance attribute `port_names_expected`.

    """

    def __init__(self, params):
        """Constructor.

        Parameters
        ----------
        params: dict
            All parameters for the module in the form of a dictionary.

        """

        super().__init__()

        self.params = deepcopy(params)
        self.params['entropy'] = 4.5 # Kj/Kg-K

        self.port_names_expected = ['inflow', 'outflow-1', 'outflow-2']

        self.initial_time = params['start-time']
        self.end_time = params['end-time']

        self.time_step = 10.0
        self.show_time = (False, 10.0)

        self.log = logging.getLogger('cortix')
        self.efficiency = 0.8

        # Inflow phase history
        quantities = list()

        temp = Quantity(name='temp', formal_name='T_in', unit='K', value=params['coolant-temp'],
                        info='Turbine Steam Inflow Temperature', latex_name=r'$T_i$')

        quantities.append(temp)

        self.inflow_phase = Phase(time_stamp=self.initial_time, time_unit='s',
                                  quantities=quantities)

        # Outflow phase history
        quantities = list()

        temp = Quantity(name='temp', formal_name='T_o', unit='K', value=params['turbine-outflow-temp'],
                        info='Turbine Steam Outflow Temperature', latex_name=r'$T_o$')

        quantities.append(temp)

        press = Quantity(name='pressure', formal_name='P_t', unit='Pa',
                         value=params['runoff-pressure'],
                         info='Turbine Steam Outflow Pressure', latex_name=r'$P_t$')

        quantities.append(press)

        x = Quantity(name='quality', formal_name='chi_t', unit='%', value=params['turbine-chi'],
                     info='Turbine Steam Outflow Quality', latex_name=r'$\chi_t$')

        quantities.append(x)

        work = Quantity(name='power', formal_name='P_t', unit='W', value=params['turbine-work'],
                        info='Turbine Power', latex_name=r'$W_t$')

        quantities.append(work)

        flowrate = Quantity(name='flowrate', formal_name='m', unit='kg/s', value=params['initial-flowrate'], info='Mass Flowrate', latex_name=r'$/dot m$')

        quantities.append(flowrate)

        self.outflow_phase = Phase(time_stamp=self.initial_time, time_unit='s',
                                   quantities=quantities)

    def run(self, *args):

        time = self.initial_time

        while time < self.end_time + self.time_step:

            if self.show_time[0] and abs(time%self.show_time[1]-0.0) <= 1.e-1:
                msg = 'time = '+str(round(time/unit.minute, 1))
                self.log.info(msg)

            # Communicate information
            #------------------------
            self.__call_ports(time)

            # Evolve one time step
            #---------------------

            time = self.__step(time)

    def module_tester(self):
        return

    def __call_ports(self, time):

        # Interactions in the steam-outflow-1 port
        #-----------------------------------------
        # one way "to" outflow-1

        # to
        if self.get_port('outflow-1').connected_port:
            print(self.params['high_pressure_turbine'])
            message_time = self.recv('outflow-1')

            outflow_state = dict()

            steam_temp = self.outflow_phase.get_value('temp', time)
            steam_quality = self.outflow_phase.get_value('quality', time)
            steam_press = self.params['turbine_inlet_pressure']
            steam_flowrate = self.outflow_phase.get_value('flowrate', time)


            outflow_state['temp'] = steam_temp
            outflow_state['quality'] = steam_quality
            outflow_state['pressure'] = steam_press
            outflow_state['flowrate'] = steam_flowrate

            if self.params['high_pressure_turbine'] == False:
                outflow_state['flowrate'] *= 2

            else:
                outflow_state['flowrate'] *= 0.5

            self.send((message_time, outflow_state), 'outflow-1')

        # Interactions in the steam-outflow-2 port
        #-----------------------------------------
        # one way "to" outflow-2

        # to
        if self.get_port('outflow-2').connected_port:

            message_time = self.recv('outflow-2')

            outflow_state = dict()

            steam_temp = self.outflow_phase.get_value('temp', time)
            steam_quality = self.outflow_phase.get_value('quality', time)
            steam_press = self.params['turbine_outlet_pressure']
            steam_flowrate = self.outflow_phase.get_value('flowrate', time)

            outflow_state['temp'] = steam_temp
            outflow_state['quality'] = steam_quality
            outflow_state['pressure'] = steam_press
            outflow_state['flowrate'] = steam_flowrate

            if self.params['high_pressure_turbine'] == False:
                outflow_state['flowrate'] *= 2

            else:
                outflow_state['flowrate'] *= 0.5

            self.send((message_time, outflow_state), 'outflow-2')

        # Interactions in the steam-inflow port
        #----------------------------------------
        # one way "from" inflow

        # from
        if self.get_port('inflow').connected_port:

            self.send(time, 'inflow')

            (check_time, inflow_state) = self.recv('inflow')
            assert abs(check_time-time) <= 1e-6

            self.temp = inflow_state['temp']
            self.chi = inflow_state['quality']
            self.pressure = inflow_state['pressure']
            self.flowrate = inflow_state['flowrate']

            if self.params['high_pressure_turbine'] == False:
                self.flowrate = self.flowrate * 0.5

    def __step(self, time=0.0):
        r"""
        Advance the state of the turbine.

        Parameters
        ----------
        time: float
            Time in SI unit.

        Returns
        -------
        None
        """
        output = self.__turbine(time, self.temp, self.chi, self.pressure, self.flowrate, self.params)

        tmp = self.outflow_phase.get_row(time)

        time += self.time_step

        self.outflow_phase.add_row(time, tmp)

        t_runoff = output[0]

        x_runoff = output[2]

        m_runoff = output[3]

        #if x_runoff < 0 and x_runoff != -1:
            #x_runoff = 0

        w_turbine = output[1]

        self.outflow_phase.set_value('power', w_turbine, time)
        self.outflow_phase.set_value('quality', x_runoff, time)
        self.outflow_phase.set_value('temp', t_runoff, time)
        self.outflow_phase.set_value('flowrate', m_runoff, time)

        return time

    def __turbine(self, time, temp_in, chi_in, p_in, flowrate, params):
        # Expand the entering steam from whatever temperature and pressure it enters
        # at to 0.035 kpa, with 80% efficiency.

        if flowrate == 0:
            return(293.15, 0, 0, 0)

        if temp_in < steam_table._TSat_P(0.5) and self.params['high_pressure_turbine'] == False:
            t_runoff = steam_table._TSat_P(0.5)
            w_real = 0
            runoff_quality = 0
            return(t_runoff, w_real, runoff_quality)

        # vfda: access of a protected member?
        inlet_parameters = steam_table._Region4(p_in, chi_in)
        inlet_entropy = inlet_parameters['s']
        inlet_enthalpy = inlet_parameters['h']

        p_out = self.params['turbine_outlet_pressure']

        #bubble and dewpoint properties at turbine outlet
        # vfda: access of a protected member?
        bubl = steam_table._Region4(p_out, 0)
        # vfda: access of a protected member?
        dew = steam_table._Region4(p_out, 1)
        bubl_entropy = bubl['s']
        dew_entropy = dew['s']
        bubl_enthalpy = bubl['h']
        dew_enthalpy = dew['h']
        #print(bubl_entropy, inlet_entropy, dew_entropy, self.params['high_pressure_turbine'])

        #if the ideal runoff is two-phase mixture:
        #if inlet_entropy < dew_entropy and inlet_entropy > bubl_entropy:
        if  bubl_entropy < inlet_entropy < dew_entropy:
            x_ideal = (inlet_entropy - bubl_entropy)/(dew_entropy - bubl_entropy)
            # vfda: access of a protected member?
            h_ideal = steam_table._Region4(p_out, x_ideal)['h']

        #if the ideal runoff is a superheated steam
        #elif inlet_entropy > dew_entropy:
            # vfda: access of a protected member?
            #t_ideal = steam_table._Backward2_T_Ps(p_out, inlet_entropy)
            # vfda: access of a protected member?
            #h_ideal = steam_table._Region2(t_ideal, p_out)['h']

        #calculate the real runoff enthalpy
        w_ideal = inlet_enthalpy - h_ideal #on a per mass basis
        #assert(w_ideal > 0)
        w_real = w_ideal * self.efficiency
        h_real = inlet_enthalpy - w_ideal
        assert h_real > 0
        if w_real < 0:
            w_real = 0

        # if the real runoff is a superheated steam
        if h_real > dew_enthalpy:
            # vfda: access of a protected member?
            t_runoff = steam_table._Backward2_T_Ph(p_out, h_real)
            x_runoff = -1 # superheated steam

        #if the real runoff is a subcooled liquid
        elif h_real < bubl_enthalpy:
            # vfda: access of a protected member?
            t_runoff = steam_table._Backward1_T_Ph(p_out, h_real)
            x_runoff = 2 # subcooled liquid

        #if the real runoff is a two-phase mixture
        else:
            # saturated vapor
            x_runoff = (h_real - bubl_enthalpy)/(dew_enthalpy - bubl_enthalpy)
            # vfda: access of a protected member?
            t_runoff = steam_table._TSat_P(p_out)
        w_real = w_real * flowrate * unit.kilo

        return (t_runoff, w_real, x_runoff, flowrate)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment.
# https://cortix.org

import pickle
import logging
import numpy as np
import scipy.constants as const
from scipy.integrate import odeint
from cortix.src.module import Module
from cortix.support.phase import Phase
from cortix.support.specie import Specie
from cortix.support.quantity import Quantity

class Droplet(Module):
    '''
    Droplet Cortix module used to model very simple fluid-particle interactions.

    Notes
    -----
    Port names used in this module: `external-flow` exchanges data with any other
    module that provides information about the flow outside the droplet,
    `visualization` sends data to a visualization module.
    '''

    def __init__(self):
        '''
        Attributes
        ----------
        initial_time: float
        end_time: float
        time_step: float
        show_time: tuple
            Two-element tuple, `(bool,float)`, `True` will print to standard
            output.
        '''

        super().__init__()

        self.port_names_expected = ['external-flow','visualization']

        self.initial_time = 0.0
        self.end_time = 100
        self.time_step = 0.1
        self.show_time = (False,1*const.minute)

        self.bounce = True
        self.slip   = True

        species = list()
        quantities = list()

        self.ode_params = dict()

        # Create a drop with random diameter up within 5 and 8 mm.
        self.droplet_diameter = (np.random.random(1) * (8 - 5) + 5)[0] * const.milli
        self.ode_params['droplet-diameter'] = self.droplet_diameter
        self.ode_params['droplet-xsec-area'] = np.pi * (self.droplet_diameter/2.0)**2
        self.ode_params['gravity'] = const.g

        # Species in the liquid phase
        water = Specie(name='water', formula_name='H2O(l)', phase='liquid', \
                atoms=['2*H','O'])
        water.massCC =  0.99965 # [g/cc]
        water.massCCUnit = 'g/cc'
        water.molarCCUnit = 'mole/cc'
        species.append(water)

        droplet_mass = 4/3 * np.pi * (self.droplet_diameter/2)**3 * water.massCC * \
                const.gram / const.centi**3  # [kg]
        self.ode_params['droplet-mass'] = droplet_mass

        # Spatial position
        x_0 = np.zeros(3)
        position = Quantity(name='position', formalName='Pos.', unit='m', value=x_0)
        quantities.append(position)

        # Velocity
        v_0 = np.zeros(3)
        velocity = Quantity(name='velocity', formalName='Veloc.', unit='m/s', value=v_0)
        quantities.append(velocity)

        # Speed
        speed = Quantity(name='speed', formalName='Speed', unit='m/s', value=0.0)
        quantities.append(speed)

        # Radial position
        radial_pos = Quantity(name='radial-position', formalName='Radius', unit='m', \
                value=np.linalg.norm(x_0[0:2]))
        quantities.append(radial_pos)

        # Liquid phase 
        self.liquid_phase = Phase(self.initial_time, time_unit='s', species=species, \
                quantities=quantities)
        self.liquid_phase.SetValue('water', water.massCC, self.initial_time)

        # Domain box dimensions: LxLxH m^3 box with given H.
        # Origin of cartesian coordinate system at the bottom of the box. 
        # z coordinate pointing upwards. -L <= x <= L, -L <= y <= L, 
        self.box_half_length = 250.0 # L [m]
        self.box_height = 500.0 # H [m]

        # Random positioning of the droplet constrained to a box sub-region.
        x_0 = (2 * np.random.random(3) - np.ones(3)) * self.box_half_length / 4.0
        x_0[2] = self.box_height
        self.liquid_phase.SetValue('position', x_0, self.initial_time)

        # Droplet Initial velocity = 0 -> placed still in the flow
        self.liquid_phase.SetValue('velocity', np.array([0.0,0.0,0.0]), \
                self.initial_time)

        # Default value for the medium surrounding the droplet if data is not passed
        # through a conneted port.
        medium_mass_density = 0.1 * const.gram / const.centi**3 # [kg/m^3]
        self.ode_params['medium-mass-density'] = medium_mass_density

        medium_displaced_mass = 4/3 * np.pi * (self.droplet_diameter/2)**3 * \
                medium_mass_density # [kg]
        self.ode_params['medium-displaced-mass'] = medium_displaced_mass

        medium_dyn_viscosity = 1.81e-5 # kg/(m s)
        self.ode_params['medium-dyn-viscosity'] = medium_dyn_viscosity

    def run(self, *args):

        time = self.initial_time

        self.bottom_impact = False

        while time < self.end_time:

            # Interactions in the external-flow port
            #---------------------------------------

            position = self.liquid_phase.GetValue('position')
            self.send( (time,position), 'external-flow' )

            (check_time, velocity,fluid_props) = self.recv( 'external-flow' )

            assert abs(check_time-time) <= 1e-6
            self.ode_params['flow-velocity'] = velocity
            #medium_mass_density  = fluid_props.mass_density  # see Vortex
            #medium_dyn_viscosity = fluid_props.dyn_viscosity # see Vortex
            medium_mass_density  = fluid_props[0]
            medium_dyn_viscosity = fluid_props[1]

            self.ode_params['medium-mass-density'] = medium_mass_density

            medium_displaced_mass = 4/3 * np.pi * (self.droplet_diameter/2)**3 * \
                    medium_mass_density # [kg]
            self.ode_params['medium-displaced-mass'] = medium_displaced_mass

            self.ode_params['medium-dyn-viscosity'] = medium_dyn_viscosity

            # Interactions in the visualization port
            #---------------------------------------

            self.send( position, 'visualization' )

            # Evolve droplet state to next time stamp
            #----------------------------------------

            time = self.__step( time )

        self.send('DONE', 'visualization') # this should not be needed: TODO
        return

    def __rhs_fn(self, u_vec, t, params):
        drop_pos = u_vec[:3]
        flow_velo = params['flow-velocity']

        drop_velo = u_vec[3:]
        relative_velo = drop_velo - flow_velo
        relative_velo_mag = np.linalg.norm(relative_velo)
        area = params['droplet-xsec-area']
        diameter = params['droplet-diameter']
        dyn_visco = params['medium-dyn-viscosity']
        rho_flow = params['medium-mass-density']

        # Calculate the friction factor
        reynolds_num = rho_flow * relative_velo_mag * diameter / dyn_visco
        if reynolds_num <= 0.0:
            fric_factor = 0.0
        elif reynolds_num < 0.1:
            fric_factor = 24 / reynolds_num
        elif reynolds_num >= 0.1 and reynolds_num < 6000.0:
            fric_factor = (np.sqrt(24 / reynolds_num) + 0.5407)**2
        elif reynolds_num >= 6000:
            fric_factor = 0.44

        drag = - fric_factor * area * rho_flow * relative_velo_mag * relative_velo / 2.0

        gravity = params['gravity']
        droplet_mass = params['droplet-mass']
        medium_displaced_mass = params['medium-displaced-mass']
        buoyant_force = (droplet_mass - medium_displaced_mass) * gravity

        dt_u_0 = u_vec[3]                               #  d_t u_1 = u_4
        dt_u_3 = drag[0]/droplet_mass                   #  d_t u_4 = f_1/m
        dt_u_1 = u_vec[4]                               #  d_t u_2 = u_5
        dt_u_4 = drag[1]/droplet_mass                   #  d_t u_5 = f_2/m
        dt_u_2 = u_vec[5]                               #  d_t u_3 = u_6
        dt_u_5 = (drag[2] - buoyant_force)/droplet_mass #  d_t u_6 = f_3/m

        return [dt_u_0, dt_u_1, dt_u_2, dt_u_3, dt_u_4, dt_u_5]

    def __step(self, time=0.0):
        r'''
        ODE IVP problem:
        Given the initial data at :math:`t=0`,
        :math:`(u_1(0),u_2(0),u_3(0)) = (x_0,x_1,x_2)`,
        :math:`(u_4(0),u_5(0),u_6(0)) = (v_0,v_1,v_2) =
        (\dot{u}_1(0),\dot{u}_2(0),\dot{u}_3(0))`,
        solve :math:`\frac{\text{d}u}{\text{d}t} = f(u)` in the interval
        :math:`0\le t \le t_f`.
        When :math:`u_3(t)` is negative, bounce the droplet to a random height between
        0 and :math:`1.0\,x_0` with no velocity, and continue the time integration until
        :math:`t = t_f`.

        Parameters
        ----------
        time: float
            Time in the droplet unit of time (seconds).

        Returns
        -------
        None
        '''

        if not self.bottom_impact:

           x_0 = self.liquid_phase.GetValue('position', time)
           v_0 = self.liquid_phase.GetValue('velocity', time)
           u_vec_0 = np.concatenate((x_0,v_0))

           t_interval_sec = np.linspace(0.0, self.time_step, num=2)

           (u_vec_hist, info_dict) = odeint(self.__rhs_fn,
                                            u_vec_0, t_interval_sec,
                                            args=( self.ode_params, ),
                                            rtol=1e-4, atol=1e-8, mxstep=300,
                                            full_output=True)

           assert info_dict['message'] =='Integration successful.', \
                   'At time: %r; state: %r; message: %r; full output: %r'%\
                   (round(time,2), u_vec_0, info_dict['message'], info_dict)

           u_vec = u_vec_hist[1,:]  # solution vector at final time step

        values = self.liquid_phase.GetRow(time) # values at previous time

        time += self.time_step

        self.liquid_phase.AddRow(time, values)

        if not self.bottom_impact:

            # Ground impact with bouncing drop
            if u_vec[2] <= 0.0 and self.bounce:
                position = self.liquid_phase.GetValue('position', self.initial_time)
                bounced_position = position[2] * np.random.random(1)
                u_vec[2]  = bounced_position
                u_vec[3:] = 0.0  # zero velocity
            # Ground impact with no bouncing drop and slip velocity
            elif u_vec[2] <= 0.0 and not self.bounce and self.slip:
                u_vec[2]  = 0.0  # don't bounce
            # Ground impact with no bouncing drop and no velocity
            elif u_vec[2] <= 0.0 and not self.bounce and not self.slip:
                u_vec[2]  = 0.0  # don't bounce
                u_vec[3:] = 0.0  # zero velocity
                self.bottom_impact = True

            # Update current values
            self.liquid_phase.SetValue('position', u_vec[0:3], time)
            self.liquid_phase.SetValue('velocity', u_vec[3:], time)
            self.liquid_phase.SetValue('speed', np.linalg.norm(u_vec[3:]), time)
            self.liquid_phase.SetValue('radial-position', np.linalg.norm(u_vec[0:2]),
                    time)

        return time



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
#
# All rights reserved, see COPYRIGHT for full restrictions.
# https://github.com/dpploy/cortix/blob/master/COPYRIGHT.txt
#
# Licensed under the University of Massachusetts Lowell LICENSE:
# https://github.com/dpploy/cortix/blob/master/LICENSE.txt
'''
This FuelBucket class is a container for usage with other plant-level process modules.
It is meant to represent a fuel bucket of a metal fuel reactor.
----------
ATTENTION:
----------
This container uses Phase() for phases (cladding and fuel). Therefore user is
responsible to make the "history" of the phases consistent. See Phase() info.

Author: Valmor de Almeida dealmeidav@ornl.gov; vfda
'''
#*********************************************************************************
import os
import sys
import math
import pandas
from copy import deepcopy
#*********************************************************************************

class FuelBucket():

#*********************************************************************************
# Construction
#*********************************************************************************

    def __init__(self,
                 specs=pandas.DataFrame()
                ):

        assert isinstance(specs, pandas.DataFrame), 'oops not pandas table.'

        self.__specs = specs

        self.__solid_phase = None

        self.__fuel_phase = None
        self.__cladding_phase = None

        return

#**********************************************************************************
# Public Member Functions
#**********************************************************************************

    # Start: Pre-irradiation information

    def get_name(self):

        '''
        Returns the name of the fuel bucket.

        Returns
        -------
        name: str
        '''

        return self.__specs.loc['Name', 1]
    name = property(get_name, None, None, None)

    def get_slug_type(self):

        '''
        Returns the type of slugs being stored in the bucket (inner slug or
        outer slug).

        Returns
        -------
        slug_type: str
        '''

        return self.__specs.loc['Slug type', 1]
    slug_type = property(get_slug_type, None, None, None)

    def get_n_slugs(self):

        '''
        Returns the number of fuel slugs in the bucket.

        Returns
        -------
        n_slugs: int
        '''

        return self.__get_n_slugs()
    n_slugs = property(get_n_slugs, None, None, None)

    def get_fuel_enrichment(self):

        '''
        Returns the enrichment of the fuel slugs in the bucket, in %.

        Returns
        -------
        fuel_enrichment: float
        '''

        return self.__get_fuel_enrichment()
    fuel_enrichment = property(get_fuel_enrichment, None, None, None)

    def get_fresh_u_mass(self):

        '''
        Returns the total amount of uranium in the bucket, in grams.

        Returns
        -------
        fresh_u_mass: float
        '''

        return self.__get_fresh_u_mass()
    fresh_u_mass = property(get_fresh_u_mass, None, None, None)

    def get_fresh_u238_mass(self):

        '''
        Returns the total amount of uranium-238 in the bucket, in grams.

        Returns
        -------
        fresh_u238_mass: float
        '''

        return self.__get_fresh_u238_mass()
    fresh_u238_mass = property(get_fresh_u238_mass, None, None, None)

    def get_fresh_u235_mass(self):
        '''
        Returns the total amount of uranium-235 in the bucket, in grams.

        Returns
        -------
        fresh_u235_mass: float
        '''

        return self.__GetFreshU235UMass()
    fresh_u235_mass = property(get_fresh_u235_mass, None, None, None)

    def get_cladding_mass(self):

        '''
        Returns the total mass of cladding material in the bucket, in grams.

        Returns
        -------
        cladding_mass: float
        '''

        return self.__get_cladding_mass()
    cladding_mass = property(get_cladding_mass, None, None, None)

    # End: Pre-irradiation information
    #------

    def get_slug_length(self):
        '''
        Returns the length of each slug in the fuel bucket.

        Returns
        -------
        slug_length: float
        '''

        return self.__get_slug_length()

    def set_slug_length(self, x):
        '''
        Sets the length of all slugs in the bucket to x. Used for chopping.

        Parameters
        ----------
        x: float
        '''

        self.__set_slug_length(x)
    slug_length = property(get_slug_length, set_slug_length, None, None)

    def get_outer_slug_od(self):
        '''
        Returns the outer diameter of the outer section of fuel, in cm. A fuel
        slug consists of an outer section of fuel and an inner section of fuel,
        with cladding on the outside of the slug and between the inner and
        outer sections of fuel.

        Returns
        -------
        outer_slug_od: float
        '''

        return self.__get_outer_slug_od()
    outer_slug_od = property(get_outer_slug_od, None, None, None)

    def get_outer_slug_id(self):
        '''
        Returns the inner diameter of the outer section of fuel, in cm.

        Returns
        -------
        outer_slug_id: float
        '''

        return self.__get_outer_slug_id()
    outer_slug_id = property(get_outer_slug_id, None, None, None)

    def get_inner_slug_od(self):
        '''
        Returns the outer  diameter of the inner section of fuel, in cm.

        Returns
        -------
        inner_slug_od: float
        '''

        return self.__get_inner_slug_od()
    inner_slug_od = property(get_inner_slug_od, None, None, None)

    def get_inner_slug_id(self):
        '''
        Returns the inner diameter of the inner section of fuel, in cm.

        Returns
        -------
        inner_slug_id: float
        '''

        return self.__get_inner_slug_id()
    inner_slug_id = property(get_inner_slug_id, None, None, None)

    def get_cladding_wall_thickness(self):
        '''
        Returns the thickness of the cladding wall which is on the outside of
        every fuel slug, and in between both sections of fuel, in cm.

        Returns
        -------
        cladding_wall_thickness: float
        '''

        return self.__get_cladding_wall_thickness()
    cladding_wall_thickness = property(
        get_cladding_wall_thickness, None, None, None)

    def get_cladding_end_thickness(self):
        '''
        Gets the thickness of the hemispherical cladding end caps that are
        placed on the top and bottom of the fuel slug, in cm.

        Returns
        -------
        cladding_end_thickness: float
        '''

        return self.__get_cladding_end_thickness()
    cladding_end_thickness = property(get_cladding_end_thickness, None, None, None)

    def get_slug_fuel_volume(self):
        '''
        Returns the volume of fuel present in a single fuel slug, in cm^3.

        Returns
        -------
        slug_fuel_volume: float
        '''

        return self.__get_slug_fuel_volume()
    slug_fuel_volume = property(get_slug_fuel_volume, None, None, None)

    def get_fuel_volume(self):
        '''
        Returns the total volume of fuel in the entire bucket, in cm^3.

        Returns
        -------
        fuel_volume: float
        '''

        return self.__get_fuel_volume()
    fuel_volume = property(get_fuel_volume, None, None, None)

    def get_slug_cladding_volume(self):
        '''
        Returns the volume of cladding present in a single fuel slug, in cm^3.

        Returns
        -------
        slug_cladding_volume: float
        '''

        return self.__get_slug_cladding_volume()
    slug_cladding_volume = property(get_slug_cladding_volume, None, None, None)

    def get_cladding_volume(self):
        '''
        Returns the total volume of cladding in the bucket, in cm^3.

        Returns
        -------
        cladding_volume: float
        '''

        return self.__get_cladding_volume()
    cladding_volume = property(get_cladding_volume, None, None, None)

    def get_fuel_mass(self):

        '''
        Returns the total mass of fuel in the solid phase in the bucket.

        Returns
        -------
        fuel_mass: float
        '''

        return self.__fuel_phase.GetValue('mass') # mass of the solid phase
    fuel_mass = property(get_fuel_mass, None, None, None)

    def get_fuel_mass_unit(self):
        '''
        Returns the unit that is used to measure the mass of fuel in the
        bucket.

        Returns
        -------
        fuel_mass_unit: str
        '''

        return self.__fuel_phase.GetQuantity('mass').unit # mass of the solid phase
    fuel_mass_unit = property(get_fuel_mass_unit, None, None, None)

    def get_radioactivity(self):   # radioactivity of the fuel bucket (fix me)
        '''
        Returns the radioactivity of the fuel bucket, in units of curies.

        Returns
        -------
        radioactivity: float
        '''

        return self.__fuel_phase.GetValue('radioactivity')
    radioactivity = property(get_radioactivity, None, None, None)

    def get_gamma_pwr(self):
        '''
        Returns the amount of gamma radiation given off by the fuel bucket,
        in units of watts.

        Returns
        -------
        gamma_pwr: float
        '''

        return self.__fuel_phase.GetValue('gamma')   # gamma pwr of the fuel bucket
    gamma_pwr = property(get_gamma_pwr, None, None, None)

    def get_heat_pwr(self):
        '''
        Returns the total amount of heat generated by the bucket, in units of
        watts.

        Returns
        -------
        heat_pwr: float
        '''

        return self.__fuel_phase.GetValue('heat')   # heat pwr of the fuel bucket
    heat_pwr = property(get_heat_pwr, None, None, None)

    def get_fuel_radioactivity(self):
        # radioactivity of the solid phase

        '''
        Returns the total radioactivity of the solid phase fuel, in units of
        curies.

        Returns
        -------
        fuel_radioactivity: float
        '''

        return self.__fuel_phase.GetValue('radioactivity')
    fuel_radioactivity = property(get_fuel_radioactivity, None, None, None)

    def get_fuel_phase(self):

        '''
        Returns the phase history of the fuel.

        Returns
        -------
        fuel_phase: pandas.core.frame.DataFrame
        '''

        return self.__fuel_phase

    def set_fuel_phase(self, phase):
        '''
        Sets the current fuel phase to a specified phase value.

        Parameters
        ----------
        phase: dataFrame
        '''

        self.__fuel_phase = deepcopy(phase)
    fuel_phase = property(get_fuel_phase, set_fuel_phase, None, None)

    def get_cladding_phase(self):
        '''
        Returns the phase history of the cladding.

        Returns
        -------
        cladding_phase: dataFrame
        '''

        return self.__cladding_phase

    def set_cladding_phase(self, phase):
        '''
        Set's the phase history to specific values.

        Parameters
        ----------
        phase: dataFrame
        '''

        self.__cladding_phase = deepcopy(phase)
    cladding_phase = property(get_cladding_phase, set_cladding_phase, None, None)

#**********************************************************************************
# Private helper functions (internal use: __)
#**********************************************************************************

    def __get_n_slugs(self):
        '''
        Returns the number of fuel slugs in the bucket.

        Returns
        -------
        self.__specs.loc['Number of slugs', 1]: int
        '''

        return int(self.__specs.loc['Number of slugs', 1])

    def __get_fuel_enrichment(self):
        '''
        Returns the enrichment of the fuel in the bucket, in %.

        Returns
        -------
        self.__specs.loc['Enrichment [U-235 wt%]', 1]: float
        '''

        return float(self.__specs.loc['Enrichment [U-235 wt%]', 1])

    def __get_outer_slug_fresh_u_mass(self):
        '''
        Returns the the mass of uranium present in the outer part of a fuel
        slug, in grams.

        Returns
        -------
        self.__specs.loc['U mass outer slug [kg]', 1]) * 1000: float
        '''

        return float(self.__specs.loc['U mass outer slug [kg]', 1]) * 1000.0  # [g]

    def __get_inner_slug_fresh_u_mass(self):
        '''
        Returns the mass of uranium present in the inner part of the fuel slug,
        in grams.

        Returns
        -------
        self.__specs.loc['U mass inner slug [kg]', 1]) * 1000: float
        '''

        return float(
            self.__specs.loc['U mass inner slug [kg]', 1]) * 1000.0  # [g]

    def __get_outer_slug_cladding_mass(self):
        '''
        Returns the mass of cladding present in the outer part of the slug, in
        grams.

        Returns
        -------
        self.__specs.loc['Cladding mass outer slug [kg]', 1]) * 1000: float
        '''

        return float(
            self.__specs.loc['Cladding mass outer slug [kg]', 1]) * 1000.0  # [g]

    def __get_inner_slug_cladding_mass(self):
        '''
        Returns the mass of cladding present in the inner part of the slug, in
        grams.
        '''

        return float(
            self.__specs.loc['Cladding mass inner slug [kg]', 1]) * 1000.0  # [g]

    def __get_fresh_u_mass(self):
        '''
        Returns the total amount of uranium present in the fuel bucket, in
        grams.

        Returns
        -------
        n_slugs * (uMassOuterSlug + uMassInnerSlug): float
        '''

        n_slugs = self.__get_n_slugs()
        uMassOuterSlug = self.__get_outer_slug_fresh_u_mass()
        uMassInnerSlug = self.__get_inner_slug_fresh_u_mass()
        return n_slugs * (uMassOuterSlug + uMassInnerSlug)

    def __get_fresh_u238_mass(self):
        '''
        Returns the total mass of uranium-238 present in the fuel bucket, in
        grams.

        Returns
        -------
        totalUMass * (1.0 - fuelEnrichment / 100): float
        '''

        totalUMass = self.__get_fresh_u_mass()
        fuelEnrichment = self.__get_fuel_enrichment()
        return totalUMass * (1.0 - fuelEnrichment / 100.0)

    def __get_fresh_u235_mass(self):
        '''
        Returns the total amount of uranium-235 present in the bucket in grams.

        Returns
        -------
        totalUmass * fuelEnrichment / 100: float
        '''

        totalUMass = self.__get_fresh_u_mass()
        fuelEnrichment = self.__get_fuel_enrichment()
        return totalUMass * fuelEnrichment / 100.0

    def __get_cladding_mass(self):
        '''
        Returns the total amount of cladding present in the bucket, in grams.

        Returns
        -------
        n_slugs * (cladMassOuterSlug + cladMassInnerSlug): float
        '''

        n_slugs = self.__get_n_slugs()
        cladMassOuterSlug = self.__get_outer_slug_cladding_mass()
        cladMassInnerSlug = self.__get_inner_slug_cladding_mass()
        return n_slugs * (cladMassOuterSlug + cladMassInnerSlug)

    def __get_slug_length(self):
        '''
        Returns the length of a fuel slug, in cm. Does NOT include end caps.

        Returns
        -------
        self.__specs.loc['Slug length [in]', 1]) * 2.54: float
        '''

        return float(self.__specs.loc['Slug length [in]', 1]) * 2.54  # cm

    def __set_slug_length(self, x):
        '''
        Sets the length of a fuel slug to a specified value. Used for chopping.

        Parameters
        ----------
        x: float
        '''

        self.__specs.loc['Slug length [in]', 1] = x / 2.54  # in
        return

    def __get_outer_slug_od(self):
        '''
        Returns the outer diameter of the outer fuel section of the slug in cm.

        Returns
        -------
        self.__specs.loc['Outer slug O.D. [in]', 1]) * 2.54: float
        '''

        return float(self.__specs.loc['Outer slug O.D. [in]', 1]) * 2.54  # cm

    def __get_outer_slug_id(self):
        '''
        Returns the inner diameter of the outer fuel section of the slug in cm.

        Returns
        -------
        self.__specs.loc['Outer slug I.D. [in]', 1]) * 2.54: float
        '''

        return float(self.__specs.loc['Outer slug I.D. [in]', 1]) * 2.54  # cm

    def __get_inner_slug_od(self):
        '''
        Returns the outer diameter of the inner fuel section of the slug in cm.

        Returns
        -------
        self.__specs.loc['Inner slug O.D. [in]', 1]) * 2.54: float
        '''

        return float(self.__specs.loc['Inner slug O.D. [in]', 1]) * 2.54  # cm

    def __get_inner_slug_id(self):
        '''
        Returns the inner diameter of the inner fuel section of the slug in cm.

        Returns
        -------
        self.__specs.loc['Inner slug I.D. [in]', 1]) * 2.54: float
        '''

        return float(self.__specs.loc['Inner slug I.D. [in]', 1]) * 2.54  # cm

    def __get_cladding_wall_thickness(self):
        '''
        Returns the thickness of the cladding material used on the outside of
        the fuel slug and between the inner and outer fuel sections.

        Returns
        -------
        self.__specs.loc['Cladding wall thickness [mm]', 1]) /10: float
        '''

        return float(
            self.__specs.loc['Cladding wall thickness [mm]', 1]) / 10.0  # cm

    def __get_cladding_end_thickness(self):
        '''
        Returns the thickness of the hemispherical end caps placed on either
        end of the cylindrical fuel slug, in cm.

        Returns
        -------
        self.__specs.loc['Cladding end cap thickness [mm]', 1]) / 10: float
        '''

        return float(
            self.__specs.loc['Cladding end cap thickness [mm]', 1]) / 10.0  # cm

    def __get_slug_fuel_volume(self):
        '''
        Returns the volume of fuel contained within a single fuel slug, in
        cm^3.

        Returns
        -------
        outerVolume + innerVolume: float
        '''

        slugLength = self.__get_slug_length()
        cladWallThickness = self.__get_cladding_wall_thickness()
        cladEndThickness = self.__get_cladding_end_thickness()
        fuelLength = slugLength - 2.0 * cladEndThickness
        fuelOuterSlugOuterRadius = self.__get_outer_slug_od() / 2.0 - cladWallThickness
        fuelOuterSlugInnerRadius = self.__get_outer_slug_id() / 2.0 + cladWallThickness
        outerVolume = fuelLength * math.pi * \
            (fuelOuterSlugOuterRadius**2 - fuelOuterSlugInnerRadius**2)
        fuelInnerSlugOuterRadius = self.__get_inner_slug_od() / 2.0 - cladWallThickness
        fuelInnerSlugInnerRadius = self.__get_inner_slug_id() / 2.0 + cladWallThickness
        innerVolume = fuelLength * math.pi * \
            (fuelInnerSlugOuterRadius**2 - fuelInnerSlugInnerRadius**2)
        return outerVolume + innerVolume

    def __get_slug_volume(self):
        '''
        Returns the total volume of a single fuel slug, in cm^3. Does not
        include the end caps.
        '''

        slugLength = self.__get_slug_length()
        outerSlugOuterRadius = self.__get_outer_slug_od() / 2.0
        outerSlugInnerRadius = self.__get_outer_slug_id() / 2.0
        outerVolume = slugLength * math.pi * \
            (outerSlugOuterRadius**2 - outerSlugInnerRadius**2)
        innerSlugOuterRadius = self.__get_inner_slug_od() / 2.0
        innerSlugInnerRadius = self.__get_inner_slug_id() / 2.0
        innerVolume = slugLength * math.pi * \
            (innerSlugOuterRadius**2 - innerSlugInnerRadius**2)
        return outerVolume + innerVolume

    def __get_slug_cladding_volume(self):
        '''
        Returns the total volume of cladding in a single fuel slug. Does not
        include the end caps. Given in units of cm^3.

        Returns
        -------
        self.__get_slug_volume() - self.__get_slug_fuel_volume(): float
        '''

        return self.__get_slug_volume() - self.__get_slug_fuel_volume()

    def __get_fuel_volume(self):
        '''
        Returns the total volume of fuel in the bucket, in cm^3.

        Returns
        -------
        slugFuelVolume * nFuelSlugs: float
        '''

        slugFuelVolume = self.__get_slug_fuel_volume()
        nFuelSlugs = self.__get_n_slugs()
        return slugFuelVolume * nFuelSlugs

    def __get_cladding_volume(self):
        '''
        Returns the total volume of cladding material in the bucket, in cm^3.
        Does not include end caps.

        Returns
        -------
        slugCladdingVolume * nFuelSlugs
        '''

        slugCladdingVolume = self.__get_slug_cladding_volume()
        nFuelSlugs = self.__get_n_slugs()
        return slugCladdingVolume * nFuelSlugs

    def __str__(self):
        '''
        Converts to string.
        '''

        s = '\nFuelBucket():\n\t*******\n\t specs:\n\t*******\n\t %s\n'
        t = '\t************\n\t fuel_phase:\n\t**********\n\t %s\n'
        u = '\t***************\n\t cladding_phase:\n\t*****************\n\t %s\n'
        stu = s + t + u
        return stu % (self.__specs, self.__fuel_phase, self.__cladding_phase)

    def __repr__(self):

        '''
        Converts to string.
        '''

        s = '\nFuelBucket():\n\t*******\n\t specs:\n\t*******\n\t %s\n'
        t = '\t************\n\t fuel_phase:\n\t**********\n\t %s\n'
        u = '\t****************\n\t cladding_phase:\n\t****************\n\t %s\n'
        stu = s + t + u
        return stu % (self.__specs, self.__fuel_phase, self.__cladding_phase)

#======================= end class FuelBucket  ===================================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of the Cortix toolkit environment
# https://cortix.org
import pandas
#import matplotlib
#matplotlib.use('Agg', warn=False)
import matplotlib.pyplot as plt

class Quantity:
    """
    todo: this probably should not have a "value" for the same reason as Species.
          this needs some thinking.
    well not so fast. This can be used to build a quantity with anything as a
    value. For instance a history of the quantity as a time series.

    """
    def __init__(self,
                 name = 'null-quantity-name',
                 formalName = 'null-quantity-formal-name', # deprecated
                 formal_name = 'null-quantity-formal-name',
                 latex_name = 'null-quantity-latex-name',
                 value = float(0.0),      # this can be any type
                 unit = 'null-quantity-unit',
                 info = 'null-quantity-info'
                ):

        assert isinstance(name, str), 'not a string.'
        self.__name = name

        assert isinstance(formalName, str), 'not a string.'
        self.__formalName = formalName  # deprecated
        self.__formal_name = formalName

        assert isinstance(formal_name, str), 'not a string.'
        self.__formal_name = formal_name

        assert isinstance(latex_name, str), 'not a string.'
        self.__latex_name = latex_name

        self.__value = value

        assert isinstance(name, str), 'not a string.'
        self.__unit = unit

        self.__name = name
        self.__value = value
        self.__unit = unit

        self.__info = info # info text such as technical name or other properties info

        return

    def SetName(self, n):
        '''
        Sets the name of the quantity in question to n.

        Parameters
        ----------
        n: str
        '''
        self.__name = n
    def get_name(self):
        '''
        Returns the name of the quantity.

        Returns
        -------
        name: str
        '''

        return self.__name
    name = property(get_name, SetName, None, None)

    def SetValue(self, v):
        '''
        Sets the numerical value of the quantity to v.

        Parameters
        ----------
        v: float

        '''
        self.__value = v
    def GetValue(self):
        '''
        Gets the numerical value of the quantity.

        Returns
        -------
        value: any type

        '''
        return self.__value
    value = property(GetValue, SetValue, None, None)

    def SetFormalName(self, fn):

        '''
        Sets the formal name of the property to fn.

        Parameters
        ----------
        fn: str
        '''

        self.__formalName = fn
        self.__formal_name = fn
    def GetFormalName(self):
        '''
        Returns the formal name of the quantity.

        Returns
        -------
        formalName: str
        '''

        #return self.__formalName
        return self.__formal_name
    formalName = property(GetFormalName, SetFormalName, None, None)
    formal_name = property(GetFormalName, SetFormalName, None, None)

    def set_latex_name(self, ln):

        '''
        Sets the LaTeX name of the property to ln.

        Parameters
        ----------
        ln: str
        '''

        self.__latex_name = ln
    def get_latex_name(self):
        '''
        Returns the formal name of the quantity.

        Returns
        -------
        formalName: str
        '''

        return self.__latex_name
    latex_name = property(get_latex_name, set_latex_name, None, None)

    def set_info(self, ln):

        '''
        Sets the LaTeX name of the property to ln.

        Parameters
        ----------
        ln: str
        '''

        self.__latex_name = ln
    def get_info(self):
        '''
        Returns the formal name of the quantity.

        Returns
        -------
        formalName: str
        '''

        return self.__info
    info = property(get_info, set_info, None, None)

    def SetUnit(self, f):
        '''
        Sets the units of the quantity to f (for example, density would be in
        units of g/cc.

        Parameters
        ----------
        f: str
        '''

        self.__unit = f
    def GetUnit(self):
        '''
        Returns the units of the quantity.

        Returns
        -------
        unit: str
        '''

        return self.__unit
    unit = property(GetUnit, SetUnit, None, None)

    def plot(self, x_scaling=1, y_scaling=1, title=None, x_label='x', y_label=None,
            file_name=None, same_axis=True, dpi=300):
        '''
        This will support a few possibities for data storage in the self.__value
        member.

        Pandas Series. If self.__value is a Pandas Series, plot against the index.
        However the type stored in the Series matter. Suppose it is a series
        of a `numpy` array. This must be of the same rank for every entry.
        This plot method assumes it is an iterable type of the same length for every
        entry in the series. A plot of all elements in the type against the index of
        the series will be made. The plot may have all elements in one axis or
        each element in its own axis.
        '''

        plt.clf()
        plt.cla()
        plt.close()

        if not isinstance(self.__value, pandas.core.series.Series):
            return
        if len(self.__value) == 1:
            return

        if not title:
            title = self.info
        if not y_label:
            if self.latex_name != 'null-quantity-latex-name':
                y_label = self.latex_name
            elif self.formal_name != 'null-quantity-formal-name':
                y_label = self.formal_name
            elif self.name != 'null-quantity-name':
                y_label = self.name
            else:
                assert False

        if isinstance(self.__value[0],float) or isinstance(self.__value[0],int) \
                or isinstance(self.__value[0],bool):
            n_dim = 1
            # Turn series of values into a series of a list of one value to allow for
            # the indexing below
            for i in range (len(self.__value[:])):
                self.__value.iat[i] = [ self.__value.iat[i] ]  # list of one element
        else:
           n_dim = len(self.__value[0])

        x = [i*x_scaling for i in self.__value.index]
        #x = self.__value.index # potential bug in matplotlib

        if same_axis:
            fig = plt.figure(self.__formal_name)
        for i in range(n_dim):
            if not same_axis:
                fig = plt.figure(self.__formal_name+str(i))
            y = list()
            for j in range(len(x)):
                y.append( self.__value.iat[j][i] ) # must use iat()

            y = [k*y_scaling for k in y]

            plt.xlabel(x_label)
            plt.ylabel(y_label)

            plt.title(title)

            plt.plot(x, y)

            if not same_axis and file_name:
                plt.savefig(file_name+str(i)+'.png', dpi=dpi)

        if same_axis and file_name:
            plt.savefig(file_name+'.png',dpi=dpi)

        return

    def __str__(self):
        '''
        Used to print the data stored by the quantity class. Will print out
        name, formal name, the value of the quantity and its unit.

        Returns
        -------
        s: str
        '''

        s = '\n\t Quantity(): \n\t name=%s; formal name=%s; latex name=%s; info=%s; value=%s[%s]'
        return s % (self.name, self.formal_name, self.latex_name, self.info, self.value, self.unit)

    def __repr__(self):
        '''
        Used to print the data stored by the quantity class. Will print out
        name, formal name, the value of the quantity and its unit.

        Returns
        -------
        s: str
        '''

        s = '\n\t Quantity(): \n\t name=%s; formal name=%s; latex name=%s; info=%s; value=%s[%s]'
        return s % (self.name, self.formal_name, self.latex_name, self.info, self.value, self.unit)
